#include "../UberLame_src/NewFix.h"
#include "../UberLame_src/CallStack.h"
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "../UberLame_src/Integer.h"
#include "../UberLame_src/Tga.h"
#include "../UberLame_src/Jpeg.h"
#include "../UberLame_src/MinMax.h"
#include "../UberLame_src/StlUtils.h"
#include "../UberLame_src/Buffer.h"
#include "../UberLame_src/Compress.h"
#include "../UberLame_src/Base64.h"

bool ReadFile(std::string &r_s_output, const char *p_s_filename)
{
	FILE *p_fr;
	if(!(p_fr = fopen(p_s_filename, "rb")))
		return false;
	int n_length;
	if(fseek(p_fr, 0, SEEK_END) ||
	   (n_length = ftell(p_fr)) < 0 ||
	   fseek(p_fr, 0, SEEK_SET)) {
		fclose(p_fr);
		return false;
	}
	r_s_output.erase();
	if(!stl_ut::Reserve_N(r_s_output, n_length)) {
		fclose(p_fr);
		return false;
	}
	r_s_output.resize(n_length);
	if(fread(&r_s_output[0], n_length, sizeof(char), p_fr) != 1) {
		fclose(p_fr);
		return false;
	}
	fclose(p_fr);
	return true;
}

#define for if(0) {} else for

inline int n_RGB_Dot(uint32_t n_value, float f_r_weight, float f_g_weight, float f_b_weight, int n_off = 0)
{
	return max(0, min(int(f_r_weight * (n_value & 0xff) +
		f_g_weight * ((n_value >> 8) & 0xff) + f_b_weight * ((n_value >> 16) & 0xff)) + n_off, 255));
}

void ImagePlane_CropBits(uint8_t *p_buffer, int n_width, int n_height, int n_bpp)
{
	_ASSERTE(n_bpp <= 8);
	if(n_bpp == 8)
		return;
	int n_shift = 8 - n_bpp;
	int n_low_bits = ((1 << n_shift) - 1);
	uint8_t n_mask = ~uint8_t(n_low_bits);
	for(uint8_t *p_end = p_buffer + n_width * n_height; p_buffer != p_end; ++ p_buffer) {
		uint8_t n_result = *p_buffer & n_mask;
		// apply masking ... but now the range reduced

		n_result |= *p_buffer >> n_bpp;
		// put something to the low bits so black is black and white is white,
		// but this "postfix" must not add on entropy (this doesn't)

		*p_buffer = n_result;
		// store
	}
}

enum {
	pred_Up,
	pred_Left,
	pred_Avg,
	pred_Paeth
};

#include "Decompress.h"

bool Predict_ImagePlane(uint8_t *p_buffer, int n_width, int n_height, int n_pred_type)
{
	int n_first_row = (n_pred_type == pred_Left)? 0 : 1;
	int n_first_col = (n_pred_type == pred_Up)? 0 : 1;
	// some predictors can't predict first line / first column

	TBuffer t_tmp;
	if(!t_tmp.Resize((n_first_row + 1) * n_width * sizeof(uint8_t)))
		return false;
	// alloc temp buffer to hold a single scanline (two scanlines if
	// predictor references previous scanline)

	for(int y = n_first_row; y < n_height; ++ y) {
		uint8_t *p_tmp = t_tmp.p_Data() + ((n_first_row)? (y % 2) * n_width : 0);
		// get scanline pointer

		if(n_first_col)
			p_tmp[0] = p_buffer[y * n_width];
		// remember first pixel in case we don't predict it, so p_tmp contains the whole scanline

		for(int x = n_first_col; x < n_width; ++ x) {
			switch(n_pred_type) {
			case pred_Up:
				p_tmp[x] = p_buffer[x + y * n_width] - p_buffer[x + (y - 1) * n_width];
				break;
			case pred_Left:
				p_tmp[x] = p_buffer[x + y * n_width] - p_buffer[(x - 1) + y * n_width];
				break;
			case pred_Avg:
				p_tmp[x] = p_buffer[x + y * n_width] - (int(p_buffer[x + (y - 1) * n_width]) +
					int(p_buffer[(x - 1) + y * n_width]) +
					int(p_buffer[(x - 1) + (y - 1) * n_width])) / 3;
				break;
			case pred_Paeth:
				{
					uint8_t b = p_buffer[x + (y - 1) * n_width], // above
						a = p_buffer[(x - 1) + y * n_width], // left
						c = p_buffer[(x - 1) + (y - 1) * n_width]; // upper left
					// get upper-left neighbourhood

					int p = int(a) + int(b) - int(c);
					int pa = abs(p - a);
					int pb = abs(p - b);
					int pc = abs(p - c);
					// estimate distances

					if(pa <= pb && pa <= pc)
						p_tmp[x] = p_buffer[x + y * n_width] - uint8_t(a);
					else if(pb <= pc)
						p_tmp[x] = p_buffer[x + y * n_width] - uint8_t(b);
					else
						p_tmp[x] = p_buffer[x + y * n_width] - uint8_t(c);
					// return nearest
				}
				break;
			}
		}
		// predict one scanline

		if(!n_first_row)
			memcpy(p_buffer + y * n_width, p_tmp, n_width * sizeof(uint8_t));
		else if(y > 1) {
			memcpy(p_buffer + (y - 1) * n_width, t_tmp.p_Data() + (1 - y % 2) * n_width,
				n_width * sizeof(uint8_t));
		}
		// copy scanline back (delayed in case predictor references previous scanline)
	}
	// predict the image

	if(n_first_row) {
		int y = n_height;
		memcpy(p_buffer + (y - 1) * n_width, t_tmp.p_Data() + (1 - y % 2) * n_width,
			n_width * sizeof(uint8_t));
	}
	// copy last scanline back

	return true;
}

int main(int n_arg_num, const char **p_arg_list)
{
	if(n_arg_num < 3) {
		fprintf(stderr, "error: missing program arguments.\n"
			"usage:\n"
			"ImageToHeader <image-file> <header-file> [--verbose] [--rle]\n"
			"\t[--rle-decoder] [--rgb|--gray]\n"
			"where image-file is source file, containing image in jpeg or tga format,\n"
			"and header-file is output file to contain image.\n"
			"--verbose parameter turns on verbose output\n"
			"--subsample n parameter sets croma subsampling to n pixels (default 2)\n"
			"--chroma-bpp m parameter sets chroma bpp to 2 * m bits (per Cb and Cr)\n"
			"--gray forces grayscale colors (single channel)\n");
		return 0;
	}

	bool b_verbose = false;
	bool b_gray = false;
	int n_subsample = 2;
	int n_chroma_bits = 6;
	// config

	const char *p_s_infile = p_arg_list[1];
	const char *p_s_outfile = p_arg_list[2];

	for(int i = 3; i < n_arg_num; ++ i) {
		if(!strcmp(p_arg_list[i], "--verbose"))
			b_verbose = true;
		else if(!strcmp(p_arg_list[i], "--gray"))
			b_gray = true;
		else if(!strcmp(p_arg_list[i], "--chroma-bpp")) {
			if(++ i == n_arg_num) {
				fprintf(stderr, "warning: parameter: \'%s\': "
					"needs value (skipping)\n", p_arg_list[i - 1]);
				break;
			}
			n_chroma_bits = atol(p_arg_list[i]);
		} else if(!strcmp(p_arg_list[i], "--subsample")) {
			if(++ i == n_arg_num) {
				fprintf(stderr, "warning: parameter: \'%s\': "
					"needs value (skipping)\n", p_arg_list[i - 1]);
				break;
			}
			n_subsample = atol(p_arg_list[i]);
		} else
			fprintf(stderr, "warning: unknown parameter: \'%s\' (skipping)\n", p_arg_list[i]);
	}
	// parse command-line

	TBmp *p_image = 0;
	if(strrchr(p_s_infile, '.') && !_stricmp(strrchr(p_s_infile, '.'), ".tga")) {
		if(b_verbose)
			printf("loading image in TGA format ...\n");
		p_image = CTgaCodec::p_Load_TGA(p_s_infile);
	} else if(strrchr(p_s_infile, '.') && (!_stricmp(strrchr(p_s_infile, '.'), ".jpg") ||
	   !_stricmp(strrchr(p_s_infile, '.'), ".jpeg"))) {
		if(b_verbose)
			printf("loading image in JPEG format ...\n");
		FILE *p_fr;
		if(p_fr = fopen(p_s_infile, "rb")) {
			p_image = CTinyJpegDecoder().p_Decode_JPEG(p_fr);
			fclose(p_fr);
		}
	} else {
		fprintf(stderr, "error: unknown input file format\n");
		return -1;
	}
	if(!p_image) {
		fprintf(stderr, "error: error loading image file \'%s\'\n", p_s_infile);
		return -1;
	}
	// load image

	if(b_gray && !p_image->b_grayscale) {
		for(uint32_t *p_pixel = p_image->p_buffer,
		   *p_end = p_image->p_buffer + (p_image->n_width * p_image->n_height);
		   p_pixel != p_end; ++ p_pixel) {
			int n_value = n_RGB_Dot(*p_pixel, 0.299f, 0.587f, 0.114f);
			*p_pixel = 0xff000000 | (n_value << 16) | (n_value << 8) | n_value;
		}
		p_image->b_grayscale = true;
	}
	b_gray = p_image->b_grayscale;
	// convert to grayscale

	if(b_verbose)
		printf("compressing image data ...\n");

	TBuffer t_packed;
	int n_afterhuff_size;
	int n_predictor_y, n_predictor_uv; // predictors chosen based onm compression
	for(int n_predictors = 0; n_predictors < 16; ++ n_predictors) {
		int n_pred_y = n_predictors / 4;
		int n_pred_uv = n_predictors % 4;
		// try all predictors

		TBuffer t_decor_data;
		if(!b_gray) {
			size_t n_image_size = p_image->n_width * p_image->n_height;
			// get image size in pixels

			int n_chroma_w = (p_image->n_width + n_subsample - 1) / n_subsample;
			int n_chroma_h = (p_image->n_height + n_subsample - 1)/ n_subsample;
			size_t n_chroma_plane_size = n_chroma_w * n_chroma_h;
			size_t n_chroma_size = n_chroma_plane_size * 2; // U, V
			// get chroma size (subsampling)

			if(!t_decor_data.Resize((n_image_size + n_chroma_size) * sizeof(uint8_t))) {
				fprintf(stderr, "error: not enough memory\n");
				return -1;
			}
			// alloc buffer

			{
				uint8_t *p_Y = t_decor_data.p_Data();
				uint8_t *p_U = p_Y + n_image_size;
				uint8_t *p_V = p_U + n_chroma_plane_size;
				// get destination pointers

				for(const uint32_t *p_src = p_image->p_buffer,
				   *p_end = p_image->p_buffer + n_image_size; p_src != p_end;)
					*p_Y ++ = n_RGB_Dot(*p_src ++, 0.299f, 0.587f, 0.114f); // calculate luminance
				_ASSERTE(p_Y == p_U);
				// get luma

				const uint32_t *p_src = p_image->p_buffer;
				for(int y = 0, h = p_image->n_height; y < h; y += n_subsample) {
					for(int x = 0, w = p_image->n_width; x < w; x += n_subsample) {
#define WEIGHTED_UV
#ifdef WEIGHTED_UV
						int u = 0, v = 0, w_sum = 0;
						for(int dy = 0; dy < n_subsample; ++ dy) {
							for(int dx = 0; dx < n_subsample; ++ dx) {
								int n_index = min(x + dx, w - 1) + min(y + dy, h - 1) * w;
								int w = 129 - abs(127 - t_decor_data.p_Data()[n_index]);
								w *= w; // weight based on lightness
								u += w * n_RGB_Dot(p_src[n_index], -0.168736f, -0.331264f, 0.5f, 128);
								v += w * n_RGB_Dot(p_src[n_index], 0.5f, -0.418688f, -0.081312f, 128);
								w_sum += w;
							}
						}
						// calculate chrominance

						*p_U ++ = uint8_t(u / w_sum);
						*p_V ++ = uint8_t(v / w_sum);
						// average
#endif //WEIGHTED_UV
#ifdef AVERAGE_UV
						int u = 0, v = 0;
						for(int dy = 0; dy < n_subsample; ++ dy) {
							for(int dx = 0; dx < n_subsample; ++ dx) {
								int n_index = min(x + dx, w - 1) + min(y + dy, h - 1) * w;
								u += n_RGB_Dot(p_src[n_index], -0.168736f, -0.331264f, 0.5f, 128);
								v += n_RGB_Dot(p_src[n_index], 0.5f, -0.418688f, -0.081312f, 128);
							}
						}
						// calculate chrominance

						*p_U ++ = uint8_t(u / (n_subsample * n_subsample));
						*p_V ++ = uint8_t(v / (n_subsample * n_subsample));
						// average
#endif //AVERAGE_UV
#ifdef MEDIAN_UV
						std::vector<uint8_t> u_list, v_list;
						if(!stl_ut::Reserve_N(u_list, n_subsample * n_subsample) ||
						   !stl_ut::Reserve_N(v_list, n_subsample * n_subsample)) {
							fprintf(stderr, "error: not enough memory\n");
							return -1;
						}
						for(int dy = 0; dy < n_subsample; ++ dy) {
							for(int dx = 0; dx < n_subsample; ++ dx) {
								int n_index = min(x + dx, w - 1) + min(y + dy, h - 1) * w;
								u_list.push_back(n_RGB_Dot(p_src[n_index], -0.168736f, -0.331264f, 0.5f, 128));
								v_list.push_back(n_RGB_Dot(p_src[n_index], 0.5f, -0.418688f, -0.081312f, 128));
							}
						}
						// calculate chrominance

						std::sort(u_list.begin(), u_list.end());
						std::sort(v_list.begin(), v_list.end());
						*p_U ++ = u_list[u_list.size() / 2];
						*p_V ++ = v_list[v_list.size() / 2];
						// median
#endif //MEDIAN_UV
					}
				}
				_ASSERTE(p_U < p_V && p_V == t_decor_data.p_Data() + t_decor_data.n_Size());
				// get chromas
			}

			{
				uint8_t *p_Y = t_decor_data.p_Data();
				uint8_t *p_U = p_Y + n_image_size;
				uint8_t *p_V = p_U + n_chroma_plane_size;
				// get destination pointers

				ImagePlane_CropBits(p_U, n_chroma_w, n_chroma_h, n_chroma_bits);
				ImagePlane_CropBits(p_V, n_chroma_w, n_chroma_h, n_chroma_bits);
				// drop some bits from chroma plane

				Predict_ImagePlane(p_Y, p_image->n_width, p_image->n_height, n_pred_y);
				Predict_ImagePlane(p_U, n_chroma_w, n_chroma_h, n_pred_uv);
				Predict_ImagePlane(p_V, n_chroma_w, n_chroma_h, n_pred_uv);
				// apply predictor
			}
		} else {
			size_t n_image_size = p_image->n_width * p_image->n_height;
			// get image size in pixels

			if(!t_decor_data.Resize(n_image_size * sizeof(uint8_t))) {
				fprintf(stderr, "error: not enough memory\n");
				return -1;
			}
			// alloc buffer

			const uint32_t *p_pixel = p_image->p_buffer;
			for(uint8_t *p_dest = t_decor_data.p_Data(), *p_end = t_decor_data.p_Data() +
			   n_image_size; p_dest != p_end; ++ p_dest, ++ p_pixel)
				*p_dest = uint8_t(*p_pixel);
			// just copy data

			{
				uint8_t *p_Y = t_decor_data.p_Data();
				Predict_ImagePlane(p_Y, p_image->n_width, p_image->n_height, n_pred_y);
				// apply predictor
			}
		}
		// decorrelate image components (if rgb, for grey it's just copy image components)

		TBuffer t_packed_data;
		if(!CRunLengthCodec::Encode(t_decor_data, t_packed_data)) {
			fprintf(stderr, "error: not enough memory\n");
			return -1;
		}
		// pack using simple RLE (easy to unpack)

		int n_afterhuff = t_packed_data.n_Size();

		{
			TBuffer t_tmp;
			if(!CHuffmanCodec::Encode(t_packed_data, t_tmp)) {
				fprintf(stderr, "error: not enough memory\n");
				return -1;
			}
			t_packed_data.Swap(t_tmp);
		}
		// huffman step

		if(!t_packed.n_Size() || t_packed.n_Size() > t_packed_data.n_Size()) {
			t_packed.Swap(t_packed_data); // save compressed data
			n_predictor_y = n_pred_y;
			n_predictor_uv = n_pred_uv; // remember predictors
			n_afterhuff_size = n_afterhuff; // remember size needed to extract huffman
		}
		// choose best of predictors
	}

	std::string s_base64;
	if(!CModifiedBase64::Encode(s_base64, t_packed.n_Size() - sizeof(uint32_t),
	   t_packed.p_Data() + sizeof(uint32_t))) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	// encode as base64, skip huffman unpack length, we save it otherwise

	if(b_verbose) {
		const char *p_pred_name_list[] = {"up", "left", "avg", "paeth"};
		printf("\t(%d bytes, using %s for predicting Y and %s for UV)\n",
			s_base64.length(), p_pred_name_list[n_predictor_y],
			p_pred_name_list[n_predictor_uv]);
		printf("\tcompression ratio: %f:1\n", float(p_image->n_width *
			p_image->n_height * ((b_gray)? 1 : 3)) / s_base64.length());

		printf("writing image data ...\n");
	}

	{
		std::string s_ident, s_header;
		if(!stl_ut::AssignCStr(s_ident, p_s_infile) ||
		   !stl_ut::AssignCStr(s_header, p_s_infile)) {
			fprintf(stderr, "error: not enough memory\n");
			return -1;
		}
		for(int i = 0; i < s_ident.length(); ++ i) {
			if(isalpha(s_ident[i]))
				s_ident[i] = toupper(s_ident[i]);
			else if(!isdigit(s_ident[i])) {
				s_ident[i] = '_';
				s_header[i] = '_';
			}
		}
		// create class name

		FILE *p_fw;
		if(!(p_fw = fopen(p_s_outfile, "w"))) {
			fprintf(stderr, "error: failed to open \'%s\' for writing\n", p_s_outfile);
			return -1;
		}

		fprintf(p_fw, "/*\n");
		fprintf(p_fw, " *\t%s\n", p_s_outfile);
		fprintf(p_fw, " *\t\t- image in header, generated from \'%s\'\n", p_s_infile);
		fprintf(p_fw, " *\t\t- using prediction / RLE / Huff / BASE64 pipe\n");
		fprintf(p_fw, " */\n\n");
		fprintf(p_fw, "#ifndef __%s_INCLUDED\n", s_ident.c_str());
		fprintf(p_fw, "#define __%s_INCLUDED\n\n", s_ident.c_str());

		fprintf(p_fw, "class C%s {\n", s_header.c_str());
		fprintf(p_fw, "public:\n");
		fprintf(p_fw, "\tstatic inline int n_Width()\n\t{\n\t\treturn %d;\n\t}\n\n", p_image->n_width);
		fprintf(p_fw, "\tstatic inline int n_Height()\n\t{\n\t\treturn %d;\n\t}\n\n", p_image->n_height);

		fprintf(p_fw, "\tstatic inline const char *p_s_Base64_Data()\n\t{\n\t\tstatic const char *p_s_data = ");
		const char *p_s_base64 = s_base64.c_str();
		for(bool b_first_line = true; *p_s_base64;) {
			fprintf(p_fw, "\t\t\t\"" + ((b_first_line)? 3 : 0));
			int n_line_len = (b_first_line)? 54 : 80;
			if(strlen(p_s_base64) > n_line_len) {
				b_first_line = false;
				for(int i = 0; i < n_line_len; ++ i)
					fputc(*p_s_base64 ++, p_fw);
				fprintf(p_fw, "\"\n");
			} else {
				while(*p_s_base64)
					fputc(*p_s_base64 ++, p_fw);
				fprintf(p_fw, "\"");
			}
		}
		fprintf(p_fw, ";\n\t\treturn p_s_data;\n\t}\n\n");

		fprintf(p_fw, "\tstatic inline uint32_t *p_Decompress()\n\t{\n\t\treturn "
			"p_Decode_RLE_Huff_Base64(p_s_Base64_Data(), %d, %d, %d, %d, %d, %d, %s);\n\t}\n\n",
			p_image->n_width, p_image->n_height, n_afterhuff_size,
			n_predictor_y, n_predictor_uv, n_subsample, (b_gray)? "true" : "false");

#if 1
		std::string s_decoder;
		if(!ReadFile(s_decoder, "Decompress.h")) {
			fprintf(stderr, "warning: failed to read \'Decompress.h\',"
				" unable to include decompress routines\n");
		} else {
			fprintf(p_fw, "protected:\n");
			int n = 0;
			while((n = s_decoder.find("\r\n", n)) != std::string::npos)
				s_decoder.erase(n, 1);
			fprintf(p_fw, "%s", s_decoder.c_str());
			// include decompress routines
		}
#endif

		fprintf(p_fw, "};\n");

		fprintf(p_fw, "\n#endif //__%s_INCLUDED\n", s_ident.c_str());

		if(ferror(p_fw)) {
			fprintf(stderr, "error: failed to write code to \'%s\'\n", p_s_outfile);
			return -1;
		}

		fclose(p_fw);
	}
	// write C++ header file

	/*{
		uint32_t *p_raster = p_Decode_RLE_Huff_Base64(s_base64.c_str(),
			p_image->n_width, p_image->n_height, n_afterhuff_size,
			n_predictor_y, n_predictor_uv, n_subsample, b_gray);
		// call decoder

		uint32_t *p_buffer_orig = p_image->p_buffer;
		{
			p_image->p_buffer = p_raster;

			if(!CTgaCodec::Save_TGA("test_decompress.tga", *p_image)) {
				fprintf(stderr, "error: failed to save tga\n");
				return -1;
			}

			delete[] p_raster;
		}
		p_image->p_buffer = p_buffer_orig;
	}*/
	// test inverse coding procedure

	if(b_verbose)
		printf("\ndone\n");

	delete[] p_image->p_buffer;
	delete p_image;

	return 0;
}
